"""
This is a variant of Aztec Dimaond

https://en.wikipedia.org/wiki/Aztec_diamond

"""

from pycsp3 import *

n = data


def n_dominos(k):
    return 6 if k == 2 else n_dominos(k - 1) + 2 * k


nDominos = n_dominos(n)

T3 = [(k, k, ANY) for k in range(nDominos)] + [(k, ANY, k) for k in range(nDominos)]
T5 = [(k, k, ANY, ANY, ANY) for k in range(nDominos)] + [(k, ANY, k, ANY, ANY) for k in range(nDominos)] + \
     [(k, ANY, ANY, k, ANY) for k in range(nDominos)] + [(k, ANY, ANY, ANY, k) for k in range(nDominos)]


def valid(i, j):
    if i < 0 or i >= n * 2 or j < 0 or j >= n * 2:
        return False
    if i < n - 1 and (j < n - 1 - i or j > n + i):
        return False
    if i > n and (j < i - n or j > 3 * n - i - 1):
        return False
    return True


def top_left(i, j):
    return valid(i, j) and not valid(i, j - 1) and not valid(i - 1, j)


def top_right(i, j):
    return valid(i, j) and not valid(i, j + 1) and not valid(i - 1, j)


def bot_left(i, j):
    return valid(i, j) and not valid(i, j - 1) and not valid(i + 1, j)


def bot_right(i, j):
    return valid(i, j) and not valid(i, j + 1) and not valid(i + 1, j)


# x[i][j] is the number of the domino in cell (i,j)
x = VarArray(size=[2 * n, 2 * n], dom=lambda i, j: range(nDominos) if valid(i, j) else None)

m = [[x[i][j]] + [x[k][q] for k, q in [(i, j - 1), (i, j + 1), (i - 1, j), (i + 1, j)] if valid(k, q)] for i in range(2 * n) for j in range(2 * n) if
     valid(i, j)]
scp3, scp5 = [t for t in m if len(t) == 3], [t for t in m if len(t) == 5]

satisfy(
    # ensuring valid positioning of dominos (over sequences of length 3)
    [scp in T3 for scp in scp3],

    # ensuring valid positioning of dominos (over sequences of length 5)
    [scp in T5 for scp in scp5],
)

if not variant():
    satisfy(
        # ensuring the two pieces of each domino occurs twice
        Cardinality(x, occurrences={k: 2 for k in range(nDominos)})
    )

    minimize(
        Sum(x[i][j] * (abs(n - i) * abs(n - j)) for i in range(2 * n) for j in range(2 * n) if valid(i, j))
    )

elif variant("mini"):
    flat_x = flatten(x)
    nf = len(flat_x)

    satisfy(
        [flat_x in [tuple(v if j == i else ANY for j in range(nf)) for i in range(nf)] for v in range(nDominos)]
    )

    minimize(
        Sum(x[n - n // 2:n + n // 2, n - n // 2:n + n // 2])
    )

# symmetry-breaking : [Increasing(x[i]) for i in range(2 * n)] [Increasing(x[:, i]) for i in range(2 * n)],
# but not all symmtries

for i in range(n * 2):
    print(
        ['tl' if top_left(i, j) else 'tr' if top_right(i, j) else 'bl' if bot_left(i, j) else 'br' if bot_right(i, j) else '*' if valid(i, j) else ' ' for j in
         range(n * 2)])
