"""
The famous logic puzzle. See, e.g., "Sudoku as a Constraint Problem" by Helmut Simonis

## Data Example
  s13a.json

## Model
  There exists different variant.

  constraints: AllDifferent, Sum, Table

## Execution:
  python Sudoku.py -data=[number,None]
  python Sudoku.py -data=<datafile.json>
  python Sudoku.py -data=<datafile.json> -variant=table
  python Sudoku.py -data=<datafile.txt> -dataparser=Sudoku_Parser.py

## Links
 - https://en.wikipedia.org/wiki/Sudoku
 - https://www.semanticscholar.org/paper/Sudoku-as-a-Constraint-Problem-Simonis/4f069d85116ab6b4c4e6dd5f4776ad7a6170faaf

## Tags
  recreational, notebook
"""

import math

from pycsp3 import *

base = data  # n (base of the grid)
assert base >= 2
n = base * base

# x[i][j] is the value of cell with coordinates (i,j)
x = VarArray(size=[n, n], dom=range(1, n + 1))

satisfy(
    # imposing distinct values on each row and each column
    AllDifferent(x, matrix=True),

    # imposing distinct values on (main) blocks
    [AllDifferent(x[i:i + base, j:j + base]) for i in range(0, n, base) for j in range(0, n, base)],

    # imposing distinct values on shaded blocks
    [AllDifferent(x[i:i + base, j:j + base]) for i in range(1, n - base, base + 1) for j in range(1, n - base, base + 1)]
)
